@@ -19,6 +19,8 @@
 #include <QObject>
 #include <QString>
 
+#include <string>
+
 namespace linglong::cli {
 
 class CLIPrinter : public Printer
@@ -33,7 +35,9 @@ public:
 
     void printErr(const utils::error::Error &) override;
     void printPackage(const api::types::v1::PackageInfoV2 &) override;
-    void printPackages(const std::vector<api::types::v1::PackageInfoV2> &) override;
+    void printPackages(const std::vector<api::types::v1::PackageInfoDisplay> &) override;
+    void
+      printSearchResult(std::map<std::string, std::vector<api::types::v1::PackageInfoV2>>) override;
     void printPruneResult(const std::vector<api::types::v1::PackageInfoV2> &list) override;
     void printContainers(const std::vector<api::types::v1::CliContainer> &) override;
     void printReply(const api::types::v1::CommonResult &) override;
@@ -46,6 +50,7 @@ public:
     void printContent(const QStringList &filePaths) override;
     void printUpgradeList(std::vector<api::types::v1::UpgradeListResult> &) override;
     void printInspect(const api::types::v1::InspectResult &result) override;
+    void printMessage(const QString &message) override;
 };
 
 } // namespace linglong::cli
